/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _ISOTROPIC_PHASE_FUNCTION_H_
#define _ISOTROPIC_PHASE_FUNCTION_H_

#include "Media/PhaseFunction/PhaseFunction.h"

namespace PhaseFunction
{
	//-------------------------------------------------------------------------------------------------
	// Isotropic phase function. 
	template<int D>
	class Isotropic: public PhaseFunction<D>
	{
       
	public:
        Isotropic(const Real mean_time): 
			PhaseFunction(mean_time)
        {}
		
		inline Spectrum sample_direction(const VectorN<D>& wi, VectorN<D> &wo, Real &pdf)const;
		inline Spectrum operator()(const VectorN<D>& wi, const VectorN<D>& wo) const
		{ 
			if( D == 2 )
				return Spectrum(.5*M_1_PI);
			else
				return Spectrum(.25*M_1_PI);
		}

        inline Real p(const VectorN<D>& wi, const VectorN<D> &wo) const   
		{
			if( D==2 )
				return (.5*M_1_PI);
			else
				return (.25*M_1_PI);
		}
		
		//TIME RESOLVED PHASE FUNCTIONS
        Spectrum sample_direction(const VectorN<D>& wi, VectorN<D> &wo, 
                                          Real &delta_time, Real &pdf) const
		{        
			return PhaseFunction<D>::sample_direction(wi,wo,delta_time,pdf);
		}

        virtual Spectrum operator()(const VectorN<D>& wi, const VectorN<D>& wo, 
                                    Real &delta_time, Real &pdf) const
        {
			return PhaseFunction<D>::operator()(wi,wo,delta_time,pdf);
		}

		virtual Real p(const VectorN<D>& wi, const VectorN<D> &wo, const Real delta_time)const
		{	
			return PhaseFunction<D>::p(wi,wo,delta_time);
		}
		
	}; // class Isotropic


	template<>
	Spectrum Isotropic<2>::sample_direction(const VectorN<2>& wi, VectorN<2> &wo, Real &pdf)const
	{
		Sampling::circular(wo,pdf);
		return (.5*M_1_PI);
	}

	template<>
	Spectrum Isotropic<3>::sample_direction(const VectorN<3>& wi, VectorN<3> &wo, Real &pdf)const
	{
		Sampling::spherical(wo,pdf);
		return (.25*M_1_PI);;
	}

	template<>
	Real Isotropic<2>::p(const VectorN<2>& wi, const VectorN<2> &wo)const
	{
		return (.5*M_1_PI);
	}

	template<>
	Real Isotropic<3>::p(const VectorN<3>& wi, const VectorN<3> &wo)const
	{
		return (.25*M_1_PI);
	}
}; //namespace PhaseFunction

#endif //_ISOTROPIC_PHASE_FUNCTION_H_